#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wrap/atomics.h"
#include "wrap/synch.h"
#include "wrap/thread.h"

/**
 * En uppgift som ska ges till en student.
 *
 * Alltid på en av formaten:
 * - a + b = c
 * - a - b = c
 * - a * b = c
 */
struct task {
    // De två första talen.
    int a;
    int b;

    // Räknesätt.
    char operator;
};

// Skapa en uppgift. Denna funktion kommer att ta lång tid att köra i och med
// att mer avancerade uppgifter kommer att bli svårare att skapa på ett bra sätt.
struct task *task_create() {
    struct task *task = malloc(sizeof(struct task));
    const char operators[3] = { '+', '-', '*' };
    task->a = rand() % 30;
    task->b = rand() % 30;
    task->operator = operators[rand() % 3];

    // Låtsats som att det tar lite tid att skapa uppgiften...
    timer_msleep(100);
    return task;
}

// Förstör en uppgift.
void task_destroy(struct task *task) {
    free(task);
}

// Beräkna svaret på uppgiften.
int task_answer(struct task *task) {
    switch (task->operator) {
    case '+':
        return task->a + task->b;
    case '-':
        return task->a - task->b;
    case '*':
        return task->a * task->b;
    default:
        return 0;
    }
}

// Skriv ut uppgiften.
void task_print(struct task *task) {
    printf("%2d %c %2d = ?\n", task->a, task->operator, task->b);
}

// Maximalt antal studenter.
#define MAX_STUDENTS 100

/**
 * Lista över studenter som ska tentera, och vilka frågor de har fått.
 */
struct students {
    // Antal studenter.
    int num_students;

    // Antal genererade uppgifter.
    struct semaphore num_generated;

    // Array med student-id:n (en array av pekare).
    const char *names[MAX_STUDENTS];

    // Array med uppgifter. En för varje student (en array av pekare).
    struct task *tasks[MAX_STUDENTS];

    // Synkronisering av "tasks"-arrayen.
    struct lock tasks_lock[MAX_STUDENTS];
};

// Skapa en lista med studenter.
struct students *students_create(int num_students, const char *names[]) {
    struct students *students = malloc(sizeof(struct students));
    students->num_students = num_students;
    sema_init(&students->num_generated, 0);
    for (int i = 0; i < num_students; i++) {
        students->names[i] = names[i];
        students->tasks[i] = NULL;
        lock_init(&students->tasks_lock[i]);
    }
    return students;
}

// Förstör en lista av studenter.
void students_destroy(struct students *students) {
    for (int i = 0; i < students->num_students; i++) {
        if (students->tasks[i])
            task_destroy(students->tasks[i]);
    }
    free(students);
}

// Funktion som körs av en tråd som ska generera uppgifter.
void students_worker(void *data) {
    struct students *to = data;

    while (true) {
        struct task *created = task_create();

        // Hitta en student utan uppgift...
        bool added = false;
        for (int i = 0; i < to->num_students; i++) {
            lock_acquire(&to->tasks_lock[i]);
            if (to->tasks[i] == NULL) {
                to->tasks[i] = created;
                added = true;
            }
            lock_release(&to->tasks_lock[i]);

            if (added) {
                break;
            }
        }

        if (added) {
            // Om vi lyckades, meddela att vi är klara.
            sema_up(&to->num_generated);
        } else {
            // Om vi misslyckades med insättningen ska vi städa upp efter oss och avsluta tråden.
            task_destroy(created);
            return;
        }
    }
}

// Generera uppgifter till alla studenter. Startar 'num_threads' trådar, och
// meddelar användaren hur många uppgifter som är klara.
void students_generate(struct students *students, int num_threads) {
    for (int i = 0; i < num_threads; i++) {
        thread_new(students_worker, students);
    }

    // Visa förloppet av genereringen medan de andra arbetar.
    int printed = 0;
    do {
        sema_down(&students->num_generated);
        printed++;
        printf("%d av %d klara...\n", printed, students->num_students);
    } while (printed < students->num_students);
}

// Skriv ut en lista med uppgifter till studenterna.
// Vi antar att denna funktion inte anropas samtidigt som någon annan funktion som använder samma students-instans.
void students_print_tasks(struct students *students) {
    for (int i = 0; i < students->num_students; i++) {
        printf("%10s: ", students->names[i]);
        task_print(students->tasks[i]);
    }
}

// Skriv ut en lista med facit för alla uppgifterna.
// Vi antar att denna funktion inte anropas samtidigt som någon annan funktion som använder samma students-instans.
void students_print_answers(struct students *students) {
    for (int i = 0; i < students->num_students; i++) {
        printf("%10s: %3d\n", students->names[i], task_answer(students->tasks[i]));
    }
}

/**
 * Huvudprogram
 *
 * Eventuella ändringar här nedanför kommer att ignoreras vid rättning av
 * uppgiften. Du får däremot ändra något nedanför ifall du vill testa din
 * lösning.
 */

int main() {
    srand(time(NULL));

    const char *names[] = {
        "andan123", "annan123", "filst123", "danth123", "klaar123", "kasla123",
        "andan321", "annan123", "filst321", "danth321", "klaar321", "kasla321"
    };
    struct students *students = students_create(12, names);
    students_generate(students, 2);

    printf("\nUppgifter:\n");
    students_print_tasks(students);
    printf("\nFacit:\n");
    students_print_answers(students);

    students_destroy(students);

    return 0;
}
